#include ".\map.h"

// Parametric constructor.
Map::Map(int Width, int Height, int Layers)
{
	// Always set Tiledata to NULL before attempting to use it!
	Tiledata = NULL;

	// Set critical Map parameters before creation.
	this->Width = Width;
	this->Height = Height;
	this->Layers = Layers;

	// Non-critical variables, but should be set nonetheless.
	this->Xoffset = 0;
	this->Yoffset = 0;
	
	// Uhg.. Dunno if this will work. I have problems with memcpy.
	memcpy(	/*target*/ this->Name, 
			/*source*/ DEFAULT_MAP_NAME, 
			/*data size*/ sizeof(char) * MAP_NAME_SIZE);	

	// Creates the new map. 
	(Reset()==ERR_OK) ? Initialized = true : Initialized = false;

	// Tip: ALWAYS CHECK Map::Initialized after the constructor is called
	// so you will KNOW without a doubt that memory was allocated for Tiledata.
}

// The destructor
Map::~Map(void)
{
	// Free memory used by the Tile Layers
	delete [] Tiledata;
	Tiledata = NULL;
}

// Loads a properly formatted map file
long Map::Load(const char* mapFilename)
{
	// Verify that the mapFilename is actually a string and has content.	
	if(!mapFilename || mapFilename == "\0") 
		return ERR_LOAD_FILENAME;		
		
	// Attempt to open the map file.
	FILE* file;// = NULL;
	file = fopen(mapFilename, "rb");

	// Verify the file opened.
	// Also check FILE::_ptr for validity. In some instances, the file struct itself can
	// be valid while FILE::_ptr is invalid.
	if(file==NULL /*|| file->_ptr==NULL*/) 
	{
		return ERR_LOAD_FILE;
	}
	
	// Read in the Map properties first, we must close the file afterwards for a 
	// short period of time.
	fread(&Layers,	sizeof(int), 1, file);
	fread(&Width,	sizeof(int), 1, file);
	fread(&Height,	sizeof(int), 1, file);
	fclose(file);

	// Check for invalid map properties. This will verify the file format.
	if(this->Layers < 1 || Layers > MAP_MAX_LAYERS)
		return ERR_LOAD_FORMAT;
	if(this->Width < 1 || this->Width > MAP_MAX_WIDTH)
		return ERR_LOAD_FORMAT;
	if(this->Height < 1 || this->Height > MAP_MAX_HEIGHT)
		return ERR_LOAD_FORMAT;

	// Reset the Tiledata. Return any Reset errors encountered.
	if(Reset() != ERR_OK)
	{
		return ERR_LOAD_FILENAME;
	}
	
	// Re-open the map file. We again read the Map properties first, but this time
	// we can read in the Tiledata because we have successfully allocated the right
	// sized Tiledata from the last few file reads.
	file = fopen(mapFilename, "rb");
	
	// Verify the file opened.
	// Also check FILE::_ptr for validity. In some instances, the file struct itself can
	// be valid while FILE::_ptr is invalid.
	if(file==NULL /*|| file->_ptr==NULL*/) return 
		ERR_LOAD_FILE;

	// Setup for the reading of each map property. If any of these fail to read exactly
	// as much data as they are supposed to read we can safely assume the format of the
	// file we're reading is invalid.
	size_t bytes_read = 0;
	size_t bytes_to_read =	(3) + 
							(2) + 
							(sizeof(char) * MAP_NAME_SIZE);

	bytes_read =  fread(&Layers,	sizeof(int),	1,	file);	
	bytes_read += fread(&Width,		sizeof(int),	1,	file);
	bytes_read += fread(&Height,	sizeof(int),	1,	file);
	bytes_read += fread(&Xoffset,	sizeof(int),	1,	file);
	bytes_read += fread(&Yoffset,	sizeof(int),	1,	file);
	bytes_read += fread(Name,		sizeof(char),	MAP_NAME_SIZE, 
														file);

	// If any of the properties failed to read properly we obviously have a file with
	// the wrong format.
	if (bytes_read != bytes_to_read)
		return ERR_LOAD_FORMAT;

	// Otherwise we can safely load the Tiledata from the file. 
	// (with a high-degree of probability)
	fread(Tiledata,	sizeof(long),
		(this->Width * this->Height) * this->Layers,
										file);
	fclose(file);

	return ERR_OK;
}

// Saves a Map's data to a properly formatted data file.
long Map::Save(const char* mapFilename)
{
	// Verify that the mapFilename has some kind of content.
	if(!mapFilename || mapFilename == "\0") 
		return ERR_SAVE_FILENAME;		
		
	// Attempt to open the map file.
	FILE* file;// = NULL;
	file = fopen(mapFilename, "wb");

	// If the file fails to open for writting in binary, return an error message.
	// Also check FILE::_ptr for validity. In some instances, the file struct itself can
	// be valid while FILE::_ptr is invalid.
	if(file==NULL /*|| file->_ptr==NULL*/) return 
		ERR_SAVE_FILE;

	// Setup for the writting of each map property. If any of these fail to write exactly
	// as much data as they are supposed to write we can safely assume a hard drive error
	// has occured such as "hard drive full".
	size_t bytes_written = 0;
	size_t bytes_to_write =	(3) + 
							(2) + 
							(sizeof(char) * MAP_NAME_SIZE);

	// Write all of the Tiledata properties to file first.
	// NOTE: Endianess is COMPLETELY IGNORED here!!!
	bytes_written = fwrite(&Layers,		sizeof(int),	1,	file);
	bytes_written += fwrite(&Width,		sizeof(int),	1,	file);
	bytes_written += fwrite(&Height,	sizeof(int),	1,	file);
	bytes_written += fwrite(&Xoffset,	sizeof(int),	1,	file);
	bytes_written += fwrite(&Yoffset,	sizeof(int),	1,	file);
	bytes_written += fwrite(Name,		sizeof(char),	MAP_NAME_SIZE, 
															file);

	// If any of the properties failed to write exactly as much data as we expected to
	// write, then it is obvious there was a hard drive error.
	if(bytes_written != bytes_to_write)
		return ERR_SAVE_FAILED;

	// Otherwise, write all of the Tiledata at the end of the map file.
	fwrite(Tiledata,sizeof(long),
		(this->Width * this->Height) * this->Layers,
												file);
	fclose(file);
		
	return ERR_OK;
}

// Resizes a map and retains as much previous map data as possible. (Irreversible tile removal can occur!)
long Map::Resize(int Width, int Height, int Layers)
{
	// Check for range errors.
	if(this->Width < 1 || Width < 1) 
		return ERR_RESIZE_WIDTH;
	if(this->Height < 1 || Height < 1) 
		return ERR_RESIZE_HEIGHT;
	if(this->Layers < 1 || Layers < 1) 
		return ERR_RESIZE_LAYERS;

	// Store the old Width & Height so we can go through the old Tiledata and save
	// as much of the old map as possible.
	int oWidth = this->Width;
	int oHeight = this->Height;
	int oLayers = this->Layers;
	
	// Create a temporary storage variable for the Old Tiledata. Make sure the allocation
	// didn't fail.
	long* TempTiledata = NULL;
	TempTiledata = new long[ (oWidth * oHeight) * oLayers ];
	if(!TempTiledata)
		return ERR_RESIZE_ALLOCATION;

	// Copy all the data from Tiledata to TempTiledata.
	memcpy(	/*target*/ TempTiledata, 
			/*source*/ Tiledata, 
			/*data size*/ sizeof(long) * (oWidth * oHeight) * oLayers);

	// Reset ALL critical and non-critical data.
	this->Width = Width;
	this->Height = Height;
	this->Layers = Layers;
	this->Xoffset = 0;
	this->Yoffset = 0;

	// Attempt to reset the map. If an error occurs return a specific error message
	// so all variables can be taken into account when debugging.
	if(int Err = Reset() != ERR_OK)
	{
		// TODO: Return specific errors from Reset().
		return ERR_RESIZE_RESET;
	}
		
	// Determine the maximum amount of data we can keep after the resize.
	int max_x = (oWidth > this->Width) ? this->Width : oWidth;
	int max_y = (oHeight > this->Height) ? this->Height : oHeight;
	int max_l = (oLayers > this->Layers) ? this->Layers : oLayers;

	// Because Reset() set all Tiledata to -1, we only need to loop through
	// the data we were able to save from before the resize.
	for(int l=1; l <= max_l; ++l)
		for(int x=0; x < max_x; ++x)
			for(int y=0; y < max_y; ++y)
			{
				int Position = (oWidth * y) + x + (oWidth * oHeight * (l - 1));
				SetTile(x, y, l, TempTiledata[ Position ]);		
			}

	// Delete TempTileData!! (forgot to do this in the old version)
	delete [] TempTiledata;

	return ERR_OK;	
}

// Resets a map's tile data. All previous Tile data is lost.
long Map::Reset(void)
{
	// Make sure the Map dimensions are at least 1x1x1. We return a proper error message
	// if any one of the dimensions is out of range.
	if( this->Width < 1 )
		return ERR_RESET_WIDTH;
	if( this->Height < 1 ) 
		return ERR_RESET_HEIGHT;
	if( this->Layers < 1 )
		return ERR_RESET_LAYERS;

	// If the tiledata doesn't already exist we must allocate new memory for it
	// before we reset the data.
	if(Tiledata) 
		delete [] Tiledata;

	Tiledata = new long[ (this->Width * this->Height) * this->Layers ];	

	// One more quick double-check in case the allocation above had failed
	// if it has, this is a bad-bad thing. 
	if(!Tiledata) 
	{
		return ERR_RESET_ALLOCATION;
	}

	// Set every tile to -1. Layers always start a 1! ( x + (Y * width) * (l=0) = 0 
	for(int l=1; l <= this->Layers; ++l)
		for(int x=0; x < this->Width; ++x)
			for(int y=0; y < this->Height; ++y)
				SetTile(x, y, l, -1);
		
	return 0;	
}

// Gets a specific tile on a specific Layer from the Map.
long Map::GetTile(int X, int Y, int Layer)
{
	// Quickly validate the X & Y parameters passed in. if they are out of range we
	// wont even attempt to use them.
	if(X < 0 || X >= this->Width) 
		return ERR_GETTILE_X_RANGE;
	if( Y < 0 || Y >= this->Height) 
		return ERR_GETTILE_Y_RANGE;
	// Remember: The Layer range is 1..Layers, NOT 0..(Layers - 1)
	if(Layer < 1 || Layer > this->Layers) 
		return ERR_GETTILE_LAYER_RANGE;
	// Return the Tile asked for.	
	int Position = (this->Width * Y) + X + (this->Width * this->Height * (Layer - 1));
	return Tiledata[ Position ];
}

// Sets a specific tile on a specific Layer in the Map. 
long Map::SetTile(int X, int Y, int Layer, long Tile)
{	
	// Quickly validate the X & Y parameters passed in. if they are out of range we
	// wont even attempt to use them.
	if(X < 0 || X >= this->Width) 
		return ERR_SETTILE_X_RANGE;
	if( Y < 0 || Y >= this->Height) 
		return ERR_SETTILE_Y_RANGE;
	// Remember: The Layer range is 1..Layers, NOT 0..(Layers - 1)
	if(Layer < 1 || Layer > this->Layers) 
		return ERR_SETTILE_LAYER_RANGE;

	// Set the new Tiledata.	
	int Position = (this->Width * Y) + X + (this->Width * this->Height * (Layer - 1));
	Tiledata[ Position ] = Tile;

	return 0;
}


// Using the map data and a CRM32Pro Tileset the map is rendered.
// RWidth is Render Width, and RHeight is Render Height. These two parameters determine how
// much of the map will be drawn. If Layer is 0 all layers will be rendered. Otherwise Layer
// will be the only layer to use for rendering.
long Map::Render(CRM32Pro_CTile* Tileset, int RWidth, int RHeight, int Layer)
{
	// Test the Tileset. IF it's not a valid Tileset we can't use it and must
	// exit out of the render method.
	if(!Tileset) 
		return ERR_RENDER_TILESET;
	if(!Tileset->GetSurface()) 
		return ERR_RENDER_TILESET;

	// Initialize these things.
	int TileWidth = 0;
	int TileHeight = 0;
	int Xoff = this->Xoffset;
	int Yoff = this->Yoffset;
	Tileset->GetOffset(&TileWidth, &TileHeight);

	// If the Render Width and Render Height "seem" invalid, make an
	// guess-ulation on what they should be. I could make a better calculation
	// using some kind of sweet sweet modulus here.
	if(RWidth < 1)
		RWidth = CRM32Pro.screen->w / TileWidth + 1; 
	if(RHeight < 1)
		RHeight = CRM32Pro.screen->h / TileHeight + 2;

	/*
		Map Render Loop
	*/
	for(int l=1; l <= this->Layers; ++l)
	{
		// If we're not drawing all of the layers, but we've not yet reached the layer 
		// we want to render, skip this layer and go to the next one.
		if((Layer != 0) && (Layer != l))
		{
			//if (Layer > l) return ERR_OK;
			continue;
		}

		// Render the current layer. 
		for(int x=0; x < RWidth; ++x)
			for(int y=0; y < RHeight; ++y)
				{
					// Calculate the Tile we want to retrieve from the Map.
					int tx = 0;
					int ty = 0;
					tx = x + ((int)Xoff / TileWidth);
					ty = y + ((int)Yoff / TileHeight);
					
					// Check Width & Height ranges and if those checks pass, retreive
					// the Tile from Tiledata.
					int tile = 0;
					if( tx > -1 && tx < this->Width && 
						ty > -1 && ty < this->Height )
						tile = this->GetTile(tx, ty, l);
					
					// otherwise show a "debug" blank Tile. (really slows it down too)
					else
					{
#ifdef _DEBUG
						SDL_Rect rect;
						rect.x = x * TileWidth - ((int)Xoff % TileWidth);
						rect.y = y * TileHeight - ((int)Yoff % TileHeight);
						rect.w = TileWidth;
						rect.h = TileHeight;
						SDL_FillRect(CRM32Pro.screen, &rect, 0x333333);
						continue;
#endif // _DEBUG_
					}

					// if the Tile is assumed valid (>0) then attempt to render.					
					if(tile>-1)
					{
						int xpos = x * TileWidth - ((int)Xoff % TileWidth);
						int ypos = y * TileHeight - ((int)Yoff % TileHeight);
						Tileset->SetPosition(xpos, ypos);

						/*
						if the tile doesn't exist in the tileset, we need to prevent
						this particular tile from being drawn with this tileset.
						*/
						if(!Tileset->Draw(CRM32Pro.screen, tile)) // -1 If failed.
						{
							// TODO: remember this tile # to avoid drawing it again.
						}

					}
				}
	}
	return ERR_OK;
	
}

// Returns the Error received formatted as a string message
char* Map::ErrorToStr(int Err)
{	
	switch (Err)
	{
		case ERR_OK: return "Ready..."; break;
		case ERR_SETTILE_X_RANGE: return "Map::SetTile X param. is out of range."; break;
		case ERR_SETTILE_Y_RANGE: return "Map::SetTile Y param. is out of range."; break;
		case ERR_SETTILE_LAYER_RANGE: return "Map::SetTile Layer param. is out of range."; break;
		case ERR_GETTILE_X_RANGE: return "Map::GetTile X param. is out of range."; break;
		case ERR_GETTILE_Y_RANGE: return "Map::GetTile Y param. is out of range."; break;
		case ERR_GETTILE_LAYER_RANGE: return "Map::GetTile Layer param. is out of range."; break;
		case ERR_RESET_WIDTH:  break;
		case ERR_RESET_HEIGHT:  break;
		case ERR_RESET_LAYERS:  break;
		case ERR_RESET_ALLOCATION:  break;
		case ERR_RESIZE_ALLOCATION: return "Map::Resize failed to allocate mem for Tiledata."; break;
		case ERR_RESIZE_WIDTH:  return "Map::Resize Width is out of range."; break;
		case ERR_RESIZE_HEIGHT: return "Map::Resize height is out of range."; break;
		case ERR_RESIZE_LAYERS: return "Map::Resize Layers is out of range."; break;
		case ERR_RESIZE_RESET: return "Map::Resize called Reset but Reset failed."; break;
		case ERR_SAVE_FILENAME: return "Map::Save mapFilename param. is invalid."; break;
		case ERR_SAVE_FILE:  return "Map::Save::file is invalid."; break;
		case ERR_SAVE_FAILED:  return "Map::Save failed to write."; break;
		case ERR_LOAD_FILENAME: return "Map::Load mapFilename param. is invalid."; break;
		case ERR_LOAD_FILE: return "Map::Load::file is invalid."; break;
		case ERR_LOAD_FORMAT:  return "Map::Load file format loaded is invalid."; break;
		case ERR_RENDER_TILESET: return "Map::Render Tileset param. is invalid."; break;
	}

	return "Error: Unknown ERR received.";
}
